package external.wha;

import jade.content.AgentAction;
import jade.content.Concept;
import jade.content.ContentElement;
import jade.content.lang.Codec;
import jade.content.lang.Codec.CodecException;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.core.AID;
import jade.domain.FIPAAgentManagement.FailureException;
import jade.domain.FIPAAgentManagement.NotUnderstoodException;
import jade.domain.FIPANames.InteractionProtocol;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.proto.ContractNetResponder;

import java.util.Random;

import misc.AgentHelper;
import ontology.CFPRequest;
import ontology.CFPResponse;
import ontology.ConfirmationRequest;
import ontology.ConfirmationResponse;
import ontology.OfferDescription;
import ontology.OrderConfirmation;
import ontology.OrderDescription;
import external.WholesalerAgent;

/**
 * 
 * Handles responding to CFPs from OAs, preparing offers and confirming them.
 *
 */
public class OrderBehaviour extends ContractNetResponder {
	private final WholesalerAgent theAgent;
	private final Random random = new Random();

	public OrderBehaviour(WholesalerAgent theAgent) {
		super(theAgent, MessageTemplate.and(
				createMessageTemplate(InteractionProtocol.FIPA_CONTRACT_NET),
				MessageTemplate.and(MessageTemplate
						.MatchOntology(theAgent.contentOntology.getName()),
						MessageTemplate
								.MatchLanguage(theAgent.contentLanguage.getName()))));
		this.theAgent = theAgent;
	}

	protected ACLMessage handleCfp(ACLMessage cfp) {
		theAgent.addBehaviour(new OrderBehaviour(theAgent));
		ACLMessage result = cfp.createReply();
		ContentElement content = null;

		AgentHelper.debugMessage(theAgent, "got CFP from "
				+ cfp.getSender().getLocalName() + ", processing");

		try {
			try {
				content = theAgent.getContentManager().extractContent(cfp);
			} catch (OntologyException e) {
				AgentHelper.errorMessage(theAgent,
						"Error relating to message and " + "ontology");
				throw new NotUnderstoodException("(" + cfp.toString() + ")");
			} catch (Codec.CodecException e) {
				AgentHelper.errorMessage(theAgent,
						"Error Parsing the message format");
				throw new NotUnderstoodException("(" + cfp.toString() + ")");
			}

			Concept action_content;

			if (content == null || !(content instanceof Action)) {
				AgentHelper.errorMessage(theAgent,
						"Message body was not an action");
				throw new NotUnderstoodException("(" + cfp.toString() + ")");
			}

			action_content = ((Action) content).getAction();

			if (action_content == null) {
				AgentHelper.errorMessage(theAgent,
						"Message body was not present");
				throw new NotUnderstoodException("(" + cfp.toString() + ")");
			} else {
				if (action_content instanceof CFPRequest) {

					CFPRequest information_request = (CFPRequest) action_content;

					// get order details
					OrderDescription theOrder = information_request
							.getOrderDescription();
					OfferDescription myOffer = createOfferForCFP(theOrder);

					CFPResponse resp = new CFPResponse();
					if (myOffer != null) {
						// send offer
						result.setPerformative(ACLMessage.PROPOSE);
						// CFPResponse resp = new CFPResponse();
						resp.setOfferDescription(myOffer);
					} else {
						// no offer to make
						// TODO
						result.setPerformative(ACLMessage.FAILURE);
					}

					try {
						// fill contents of response
						AgentAction resp_content = resp;
						Action action = new Action(cfp.getSender(),
								resp_content);
						theAgent.getContentManager().fillContent(result, action);
					} catch (CodecException e) {
						AgentHelper.errorMessage(theAgent,
								"Error relating to message and " + "ontology");
						throw new NotUnderstoodException("("
								+ result.toString() + ")");
					} catch (OntologyException e) {
						AgentHelper.errorMessage(theAgent,
								"Error Parsing the message format");
						throw new NotUnderstoodException("("
								+ result.toString() + ")");
					}
				} else {
					AgentHelper.errorMessage(theAgent,
							"Message body was not a correct action omg");
					throw new NotUnderstoodException("(" + result.toString()
							+ ")");
				}
			}
		} catch (NotUnderstoodException e) {
			result.setPerformative(ACLMessage.NOT_UNDERSTOOD);
			result.setContent(e.getMessage());
		}
		return result;
	}

	protected ACLMessage handleAcceptProposal(ACLMessage cfp,
			ACLMessage propose, ACLMessage accept) throws FailureException {
		ACLMessage result = accept.createReply();
		ContentElement content = null;

		AgentHelper.debugMessage(theAgent, "got accept proposal from "
				+ accept.getSender().getLocalName() + ", confirming");

		try {
			try {
				content = theAgent.getContentManager().extractContent(accept);
			} catch (OntologyException e) {
				AgentHelper.errorMessage(theAgent,
						"Error relating to message and " + "ontology");
				throw new NotUnderstoodException("(" + accept.toString() + ")");
			} catch (Codec.CodecException e) {
				AgentHelper.errorMessage(theAgent,
						"Error Parsing the message format");
				throw new NotUnderstoodException("(" + accept.toString() + ")");
			}

			Concept action_content;

			if (content == null || !(content instanceof Action)) {
				AgentHelper.errorMessage(theAgent,
						"Message body was not an action");
				throw new NotUnderstoodException("(" + accept.toString() + ")");
			}

			action_content = ((Action) content).getAction();

			if (action_content == null) {
				AgentHelper.errorMessage(theAgent,
						"Message body was not present");
				throw new NotUnderstoodException("(" + accept.toString() + ")");
			} else {
				if (action_content instanceof ConfirmationRequest) {

					ConfirmationRequest information_request = (ConfirmationRequest) action_content;

					// get order details
					OfferDescription theOffer = information_request
							.getOfferDescription();
					OrderConfirmation confirm = createConfirmation(theOffer);

					ConfirmationResponse resp = new ConfirmationResponse();
					if (confirm != null) {
						// send confirmation
						result.setPerformative(ACLMessage.INFORM);
						resp.setOrderConfirmation(confirm);
					} else {
						// no confirmation
						// TODO
						result.setPerformative(ACLMessage.FAILURE);
					}

					AgentAction respContent = resp;
					Action action = new Action(new AID(), respContent);

					try {
						// fill contents of response
						theAgent.getContentManager().fillContent(result, action);
					} catch (CodecException e) {
						AgentHelper.errorMessage(theAgent,
								"Error relating to message and " + "ontology");
						throw new NotUnderstoodException("("
								+ result.toString() + ")");
					} catch (OntologyException e) {
						AgentHelper.errorMessage(theAgent,
								"Error Parsing the message format");
						throw new NotUnderstoodException("("
								+ result.toString() + ")");
					}
				} else {
					AgentHelper.errorMessage(theAgent,
							"Message body was not a correct action");
					throw new NotUnderstoodException("(" + result.toString()
							+ ")");
				}
			}
		} catch (NotUnderstoodException e) {
			result.setPerformative(ACLMessage.NOT_UNDERSTOOD);
			result.setContent(e.getMessage());
		}
		return result;
	}

	public void reset() {
		super.reset();
	}

	protected void handleNotUnderstood(ACLMessage msg) {
		AgentHelper.debugMessage(theAgent, msg.getSender().getLocalName()
				+ " did not understand my request:" + msg.getContent());
	}

	protected void handleOutOfSequence(ACLMessage msg) {
		AgentHelper.debugMessage(theAgent, msg.getSender().getLocalName()
				+ " sent me unexpected message:" + msg.getContent());
	}
	
	private OfferDescription createOfferForCFP(OrderDescription order) {
		// now agrees to everything
		OfferDescription offer = new OfferDescription();
		offer.setAgentAID(theAgent.getAID());
		offer.setAmountSpecific(order.getAmountRequired());
		offer.setDeliveryTimeSpecific(order.getDeliveryTimeRequired() / 2 - 1);
		offer.setPriceSpecific(random.nextFloat());
		offer.setTheProduct(order.getTheProduct());
		return offer;
	}
	
	private OrderConfirmation createConfirmation(OfferDescription offer) {
		// might additionally verify if the accept proposal quotes really the correct offer
		OrderConfirmation confirm = new OrderConfirmation();
		String orderId = Integer.toString(random.nextInt(1000));
		confirm.setOfferDescription(offer);
		confirm.setOrderId(orderId);
		return confirm;
	}
}