package internal.la;

import internal.LogisticsAgent;
import internal.la.FinalStateBehaviour;
import internal.la.ListBehaviour;
import internal.la.MonitorBehaviour;
import internal.la.OrderBehaviour;
import internal.la.PreOrderBehaviour;
import internal.la.ReminderBehaviour;
import jade.content.AgentAction;
import jade.content.Concept;
import jade.content.ContentElement;
import jade.content.lang.Codec;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.DataStore;
import jade.core.behaviours.FSMBehaviour;
import jade.domain.FIPAAgentManagement.NotUnderstoodException;
import jade.domain.FIPANames.InteractionProtocol;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.proto.states.MsgReceiver;
import jade.util.leap.ArrayList;
import jade.util.leap.Iterator;
import jade.util.leap.List;
import misc.AgentHelper;
import ontology.CICResponse;
import ontology.DeliveryDescription;
import ontology.IssueOrder;
import ontology.LAAgentDescription;
import ontology.OrderConfirmation;
import ontology.OrderDescription;

/**
 * Main FSM behaviour of LA.
 */
public class IssueOrderBehaviour extends FSMBehaviour {
	// state name constants
	private final static String STATE_REQUEST = "request";
	private final static String STATE_PREORDER = "preOrder";
	private final static String STATE_LIST = "getList";
	private final static String STATE_ORDER = "orderGoods";
	private final static String STATE_MONITOR = "monitorDelivery";
	private final static String STATE_REMIND = "remind";
	private final static String STATE_FINAL = "final";
	
	public final static String ORDER_REQUEST_MSG = "orderrequest";
	
	public LogisticsAgent theAgent = (LogisticsAgent) myAgent; // our agent
	public OrderDescription theOrder = null; // order to be made
	public OrderConfirmation theConfirmation = null; // confirmation received
	public ACLMessage cicQueryMessage; // message sent do CIC
	public ACLMessage issueOrderTemplate; // template message sent to OAs
	public ACLMessage issueOrderMessage; // actual message that was sent to OA
	public ACLMessage orderRequestMessage; // message received from WA
	public ACLMessage reminderMessage; // message received as response to reminder
	public int currentOA; // index of OA in the pool
	public DataStore data = new DataStore(); // data store for generic message receivers
	public ArrayList theSuppliers; // list of potential suppliers for given order
	public ArrayList involvedSuppliers; // suppliers which were actually involved in current order (for reliability)
	public int theDeadline; // overall time for completing the order request
	public int theTimeout; // time to wait for delivery from the current supplier
	public DeliveryDescription theDelivery = null; // description of received delivery

	// retrieves message from datastore by given key
	public ACLMessage retrieveData(String key) {
		return (ACLMessage) data.get(key);
	}

	// initializes FSM Behaviour
	public IssueOrderBehaviour(Agent agent) {
		super(agent);
		currentOA = 0;
		involvedSuppliers = new ArrayList();
		theSuppliers = new ArrayList();

		AgentHelper.debugMessage(this.myAgent, "configuring FSM");

		// create template for WA messages
		MessageTemplate mt = MessageTemplate.and(MessageTemplate
				.MatchPerformative(ACLMessage.REQUEST), MessageTemplate.and(
				MessageTemplate.MatchOntology(theAgent.contentOntology.getName()),
				MessageTemplate.MatchLanguage(theAgent.contentLanguage.getName())));

		// *** set FSM states
		registerFirstState(new MsgReceiver(agent, mt, MsgReceiver.INFINITE,
				data, ORDER_REQUEST_MSG), STATE_REQUEST); // listen for WA messages
		registerState(new PreOrderBehaviour(this), STATE_PREORDER); // process them
		registerState(new ListBehaviour(this, cicQueryMessage), STATE_LIST); // get product supplier list from CIC
		registerState(new OrderBehaviour(this, issueOrderTemplate), STATE_ORDER); // tell OA to order 
		registerState(new MonitorBehaviour(this), STATE_MONITOR); // wait for delivery
		registerState(new ReminderBehaviour(this, null), STATE_REMIND); // ask to remind
		registerLastState(new FinalStateBehaviour(this), STATE_FINAL); // finalize, manage reliability

		// *** set FSM transitions
		registerDefaultTransition(STATE_REQUEST, STATE_PREORDER); // always
		registerDefaultTransition(STATE_PREORDER, STATE_LIST); // always
		registerDefaultTransition(STATE_LIST, STATE_FINAL); // list not received
		registerDefaultTransition(STATE_ORDER, STATE_FINAL); // ordering failed
		registerDefaultTransition(STATE_MONITOR, STATE_FINAL); // just by default
		registerDefaultTransition(STATE_REMIND, STATE_MONITOR); // default, see if time left (used when no OA available)
		
		registerTransition(STATE_LIST, STATE_ORDER, 0); // list received
		registerTransition(STATE_LIST, STATE_FINAL, 1); // no suppliers
		registerTransition(STATE_ORDER, STATE_MONITOR, 0); // ordering successful
		registerTransition(STATE_ORDER, STATE_ORDER, 1,
				new String[] { STATE_ORDER }); // try another OA
		registerTransition(STATE_ORDER, STATE_FINAL, 2); // no OAs available
		registerTransition(STATE_MONITOR, STATE_FINAL, 0); // delivery received
		registerTransition(STATE_MONITOR, STATE_ORDER, 1, new String[] {
				STATE_ORDER, STATE_MONITOR }); // delivery not received, but still time for new order
		registerTransition(STATE_MONITOR, STATE_FINAL, 3); // delivery not received, time is up
		registerTransition(STATE_MONITOR, STATE_REMIND, 4); // delivery not received, can still remind
		registerTransition(STATE_REMIND, STATE_MONITOR, 0,
				new String[] { STATE_MONITOR }); // remind successful
		registerTransition(STATE_REMIND, STATE_REMIND, 1, new String[] { "remind" }); // try another OA
	}
	
	// sets the list of suppliers for OA using CIC query result
	public void setSupplierList(LogisticsAgent theAgent, ACLMessage cicResponse, ArrayList theSupplierList) {
		ContentElement content = null;

		try {
			try {
				content = theAgent.getContentManager().extractContent(cicResponse);
			} catch (OntologyException e) {
				AgentHelper.errorMessage(theAgent, "Error relating to message and "
						+ "ontology");
				throw new NotUnderstoodException("(" + cicResponse.toString()
						+ ")");
			} catch (Codec.CodecException e) {
				AgentHelper.errorMessage(theAgent,
						"Error Parsing the message format");
				throw new NotUnderstoodException("(" + cicResponse.toString()
						+ ")");
			}

			Concept action_content;

			if (content == null || !(content instanceof Action)) {
				AgentHelper
						.errorMessage(theAgent, "Message body was not an action");
				throw new NotUnderstoodException("(" + cicResponse.toString()
						+ ")");
			}

			action_content = ((Action) content).getAction();

			if (action_content == null) {
				AgentHelper.errorMessage(theAgent, "Message body was not present");
				throw new NotUnderstoodException("(" + cicResponse.toString()
						+ ")");
			} else {
				if (action_content instanceof CICResponse) {

					CICResponse information_request = (CICResponse) action_content;

					// get suppliers from CIC and create LAAgentDescriptions
					LAAgentDescription desc = null;
					for (Iterator iter = information_request.getAllAgentList(); iter
							.hasNext();) {
						AID wha = (AID) iter.next();
						// get reliability for supplier
						int reliability = theAgent.getReliability(wha);
						// initialize description
						desc = new LAAgentDescription();
						desc.setAgentAID(wha);
						desc.setAgentRank(reliability);
						// save it
						theSupplierList.add(desc);
					}
				} else {
					AgentHelper.errorMessage(theAgent,
							"Message body was not a correct action");
					throw new NotUnderstoodException("(" + cicResponse.toString()
							+ ")");
				}
			}
		} catch (NotUnderstoodException e) {
			;
		}
	}
	
	// used to remove supplier from the list of possible suppliers
	public void removeSupplier(IssueOrderBehaviour theMaster,
			AID chosenSupplier) {
		for (Iterator iter = theMaster.theSuppliers.iterator(); iter.hasNext();) {
			LAAgentDescription supplier = (LAAgentDescription) iter.next();
			if (supplier.getAgentAID().compareTo(chosenSupplier) == 0) {
				iter.remove();
				break;
			}
		}
		// recreate template not to include the supplier just removed
		theMaster.issueOrderTemplate = prepareIssueOrder(theMaster);
	}
	
	// prepare message template for OA
	public ACLMessage prepareIssueOrder(IssueOrderBehaviour theMaster) {
		LogisticsAgent theAgent = theMaster.theAgent;
		ACLMessage request = new ACLMessage(ACLMessage.REQUEST);
		request.setSender(theAgent.getAID());
		request.setLanguage(theAgent.contentLanguage.getName());
		request.setOntology(theAgent.contentOntology.getName());

		request.setProtocol(InteractionProtocol.FIPA_REQUEST);

		IssueOrder issue = new IssueOrder();
		// set the supplier list we got from CIC and enriched with reliability
		// data
		issue.setAgentLADescriptions((List) theMaster.theSuppliers);
		// set the order that we want to place
		issue.setOrderDescription(theMaster.theOrder);

		AgentAction content = issue;
		Action action = new Action(new AID(), content);

		try {
			AgentHelper.debugMessage(theAgent, "preparing issue order");
			// fill content of the message
			theAgent.getContentManager().fillContent(request, action);
		} catch (OntologyException e) {
			AgentHelper.errorMessage(theAgent,
					"Error relating to message and ontology");
		} catch (Codec.CodecException e) {
			AgentHelper.errorMessage(theAgent, "Error Parsing the message format");
		}
		return request;
	}
}