package internal.oa;

import internal.OrderingAgent;
import jade.core.AID;
import jade.core.behaviours.OneShotBehaviour;

import java.util.Iterator;

import misc.AgentHelper;
import ontology.LAAgentDescription;
import ontology.OfferDescription;

// ranks received proposals
public class RankingBehaviour extends OneShotBehaviour {

	// averages for normalization
	private double averagePrice;
	private double averageTime;
	private double averageAmount;
	private double averageReliability;

	private final OrderingBehaviour theMaster;
	private int result = 1;
	private final OrderingAgent theAgent;

	public RankingBehaviour(OrderingBehaviour theMaster) {
		this.theMaster = theMaster;
		theAgent = theMaster.theAgent;
	}

	// do everything
	public void action() {
		AgentHelper.debugMessage(theAgent, "ranking offers");
		calculateStrategy();
		calculateAverages();
		rankOffers();
		result = 0;
	}

	// calculates the averages
	private void calculateAverages() {
		double sumPrice = 0;
		double sumTime = 0;
		double sumAmount = 0;
		double sumReliability = 0;

		// sum up
		for (Iterator iter = theMaster.offerList.iterator(); iter.hasNext();) {
			OfferDescription offer = (OfferDescription) iter.next();
			sumPrice += offer.getPriceSpecific();
			sumTime += offer.getDeliveryTimeSpecific();
			sumAmount += offer.getAmountSpecific();
			sumReliability += getSupplierReliability(offer.getAgentAID(),
					theMaster);
		}
		// and get average
		averagePrice = sumPrice / theMaster.offerList.size();
		averageTime = sumTime / theMaster.offerList.size();
		averageAmount = sumAmount / theMaster.offerList.size();
		averageReliability = sumReliability / theMaster.offerList.size();
	}

	// calculate Strategy Factor
	private void calculateStrategy() {
		theMaster.theAgent.strategyFactor = (double) theMaster.theOrder.getDeliveryTimeAllowed()
				/ (double) theMaster.theOrder.getDeliveryTimeRequired();
	}

	// get reliability for supplier identified by given AID
	private int getSupplierReliability(AID supplier, OrderingBehaviour theMaster) {
		LAAgentDescription desc = null;
		// search for agent description
		for (Iterator iter = theMaster.theSuppliers.iterator(); iter.hasNext();) {
			desc = (LAAgentDescription) iter.next();
			if (desc.getAgentAID() == supplier)
				break;
		}
		// get rank if found
		if (desc != null)
			return desc.getAgentRank();
		else {
			// not found
			AgentHelper.errorMessage(theAgent, "reliability for agent"
					+ supplier.getLocalName() + "not found!");
			return 0;
		}
	}

	// result for FSM
	public int onEnd() {
		return result;
	}

	// SRF - more human friendly version in docs
	private double rankingFormula(double nP, double nT, double nA,
			double nR, OrderingBehaviour theMaster) {
		double wP = theMaster.theAgent.weightPrice;
		double wT = theMaster.theAgent.weightTime;
		double wA = theMaster.theAgent.weightAmount;
		double wR = theMaster.theAgent.weightReliability;
		double wS = theMaster.theAgent.weightStrategy;
		double sf = theMaster.theAgent.strategyFactor;
		double numerator = wP * nP + wT * nT + wA * nA + wR * nR;
		double denominator = wP + wT + wA + wR;
		double strategy = 2 * wS * sf * nR;
		double srf = numerator / denominator + strategy;
		return srf;
	}
	
	// ranks offers
	private void rankOffers() {
		double normPrice = 0;
		double normTime = 0;
		double normAmount = 0;
		double normReliability = 0;

		// for each offer
		for (Iterator iter = theMaster.offerList.iterator(); iter.hasNext();) {
			OfferDescription offer = (OfferDescription) iter.next();
			// normalize value
			normPrice = (offer.getPriceSpecific() - averagePrice)
					/ averagePrice;
			normTime = (offer.getDeliveryTimeSpecific() - averageTime)
					/ averageTime;
			normAmount = (offer.getAmountSpecific() - averageAmount)
					/ averageAmount;
			normReliability = (getSupplierReliability(offer.getAgentAID(),
					theMaster) - averageReliability)
					/ averageReliability;
			// save SRF in the list
			theMaster.offerRanks.add(new Double(rankingFormula(normPrice,
					normTime, normAmount, normReliability, theMaster)));
		}
	}
}
