package internal.wa;

import internal.WarehouseAgent;
import jade.content.AgentAction;
import jade.content.lang.Codec.CodecException;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.core.behaviours.CyclicBehaviour;
import jade.domain.FIPANames.InteractionProtocol;
import jade.lang.acl.ACLMessage;

import java.util.Iterator;

import misc.AgentHelper;
import ontology.OrderDescription;
import ontology.OrderRequest;
import ontology.PredictionDescription;

/**
 * 
 * Behaviour monitoring stock levels and ordering if necessary
 * 
 */
public class MonitoringBehaviour extends CyclicBehaviour {
	
	private final WarehouseAgent theAgent;

	public MonitoringBehaviour(WarehouseAgent agent) {
		super(agent);
		this.theAgent = agent;
	}

	public void onStart() {
		AgentHelper.debugMessage(myAgent, "monitoring stock levels");
	}

	public void action() {
		block(theAgent.MONITORING_INTERVAL);
		theAgent.increaseTime();
		AgentHelper.debugMessage(myAgent, "checking stock levels");

		PredictionDescription desc = null;
		OrderDescription order = null;
		int i = 0;
		// for each prediction we monitor...
		for (Iterator iter = theAgent.predictionList.iterator(); iter.hasNext(); i++) {
			desc = (PredictionDescription) iter.next();
			// ...get order, if its required
			order = createOrderForPrediction(desc, i);
			if (order != null) {
				// we got order details, so do it
				AgentHelper.debugMessage(theAgent,
						"making order request for "
								+ order.getTheProduct().getProductName());
				ACLMessage request = prepareOrderRequest(order);
				theAgent.addBehaviour(new OrderRequestBehaviour(theAgent,
						request));
			} else {
				// no need to order
			}
			order = null;
		}
	}
	
	// if it is time, prepares order request, if not, returns null, so no order
	// is placed
	private OrderDescription createOrderForPrediction(
			PredictionDescription prediction, int i) {
		OrderDescription theOrder = null;
		// is it time already?
		if (Math.abs(theAgent.getTime() - ((Integer) theAgent.lastCheckTime.get(i)).intValue()) < prediction
				.getPredictionPeriod()) {
			// NO
			return null;
		} else {
			// YES
			// check levels and create order request
			theOrder = createOrder(prediction, theAgent.getStockForProduct());
			// set last check time to now
			theAgent.lastCheckTime.set(i, new Integer(theAgent.getTime()));
		}
		return theOrder;
	}
	
	// prepares order details
	private OrderDescription createOrder(PredictionDescription prediction,
			int currentStock) {
		// order only as much as we need to satisfy predicted demand
		float safetyBuffer = prediction.getPredictionDeviation() * 2;
		float orderAmount = prediction.getPredictionAmount() - currentStock;
		OrderDescription desc = new OrderDescription();
		// preferably also fill safety buffer
		desc.setAmountPreferred(orderAmount + safetyBuffer);
		desc.setAmountRequired(orderAmount);
		desc.setDeliveryTimeAllowed(prediction.getPredictionPeriod());
		desc.setDeliveryTimeRequired(20);

		desc.setPriceMax(10);
		desc.setTheProduct(prediction.getTheProduct());
		return desc;
	}
	
	// prepares order request message
	private ACLMessage prepareOrderRequest(OrderDescription order) {
		ACLMessage msg = new ACLMessage(ACLMessage.REQUEST);
		msg.setSender(theAgent.getAID());
		msg.addReceiver(theAgent.myLA);
		msg.setLanguage(theAgent.contentLanguage.getName());
		msg.setOntology(theAgent.contentOntology.getName());
		msg.setProtocol(InteractionProtocol.FIPA_REQUEST);

		OrderRequest request = new OrderRequest();
		request.setOrderDescription(order);

		AgentAction content = request;
		Action action = new Action(theAgent.myLA, content);

		try {
			// fill contents of request
			theAgent.getContentManager().fillContent(msg, action);
		} catch (CodecException e) {
			AgentHelper.errorMessage(theAgent, "Error relating to message and "
					+ "ontology:" + e);
			return null;
		} catch (OntologyException e) {
			AgentHelper.errorMessage(theAgent, "Error Parsing the message format:"
					+ e);
			return null;
		}

		return msg;
	}
}